@@ -3,14 +3,18 @@
 #
 #  (C) Paul Evans, 2022 -- leonerd@leonerd.org.uk
 
-package Future::XS 0.02;
+package Future::XS 0.03;
 
 use v5.14;
 use warnings;
 
+use Carp;
+
 require XSLoader;
 XSLoader::load( __PACKAGE__, our $VERSION );
 
+use Time::HiRes qw( tv_interval );
+
 =head1 NAME
 
 C<Future::XS> - experimental XS implementation of C<Future>
@@ -55,6 +59,41 @@ sub import
          *{"Future::${name}"} = \&{__PACKAGE__."::${name}"};
       }
    }
+
+   croak "Unrecognised $pkg\->import symbols - " . join( ", ", sort keys %syms )
+      if %syms;
+}
+
+# These methods aren't on the "fast path" so for now we'll just implement them in Perl
+
+sub transform
+{
+   my $self = shift;
+   my %args = @_;
+
+   my $xfrm_done = $args{done};
+   my $xfrm_fail = $args{fail};
+
+   return $self->then_with_f(
+      sub {
+         my $self = shift;
+         return $self unless $xfrm_done;
+         return $self->done( $xfrm_done->( $self->result ) );
+      },
+      sub {
+         my $self = shift;
+         return $self unless $xfrm_fail;
+         return $self->fail( $xfrm_fail->( $self->failure ) );
+      },
+   );
+}
+
+sub elapsed
+{
+   my $self = shift;
+   return undef unless
+      defined( my $btime = $self->btime ) and defined( my $rtime = $self->rtime );
+   return tv_interval( $btime, $rtime );
 }
 
 =head1 AUTHOR
